import contextlib
import functools
import logging
import typing

from typing import Any, Type

import attr
import six

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.configuration.environ_config import is_databand_enabled
from dbnd._core.constants import (
    RESULT_PARAM,
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    TaskRunState,
)
from dbnd._core.current import current_task_run, get_databand_run, try_get_current_task
from dbnd._core.decorator.decorated_task import _DecoratedTask
from dbnd._core.decorator.dynamic_tasks import (
    _handle_dynamic_error,
    create_and_run_dynamic_task_safe,
)
from dbnd._core.decorator.func_task_call import FuncCall
from dbnd._core.decorator.lazy_object_proxy import CallableLazyObjectProxy
from dbnd._core.decorator.schemed_result import FuncResultParameter
from dbnd._core.decorator.task_decorator_spec import (
    _TaskDecoratorSpec,
    args_to_kwargs,
    build_task_decorator_spec,
)
from dbnd._core.errors import show_exc_info
from dbnd._core.errors.errors_utils import log_exception, user_side_code
from dbnd._core.failures import dbnd_handle_errors
from dbnd._core.plugin.dbnd_airflow_operator_plugin import (
    build_task_at_airflow_dag_context,
    is_in_airflow_dag_build_context,
)
from dbnd._core.task.task import Task
from dbnd._core.task_build.task_context import (
    TaskContextPhase,
    current_phase,
    try_get_current_task,
)
from dbnd._core.task_build.task_metaclass import TaskMetaclass
from dbnd._core.task_build.task_passport import TaskPassport
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.utils.timezone import utcnow
from targets import InMemoryTarget
from targets.values import get_value_type_of_obj


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun

logger = logging.getLogger(__name__)


@attr.s
class FuncCallWithResult(FuncCall):
    result = attr.ib(default=None)

    def set_result(self, value):
        self.result = value
        return value


class TaskClsBuilder(object):
    def __init__(self, func_spec, task_type, task_defaults):
        # type: (TaskClsBuilder, _TaskDecoratorSpec, Type[_DecoratedTask], Any) -> None
        self.func_spec = func_spec
        self.task_type = task_type
        self.task_defaults = task_defaults

        self._normal_task_cls = None
        self._tracking_task_cls = None
        # self.task_cls = task_cls  # type: Type[Task]
        # this will make class look like a origin function
        functools.update_wrapper(self, self.func)
        self._call_count = 0
        self._call_as_func = False
        self._max_call_count = get_dbnd_project_config().max_calls_per_run

        self._callable_item = None

    @property
    def func(self):
        return self.func_spec.item

    def get_task_cls(self):
        if _is_tracking_mode():
            if not self._tracking_task_cls:
                self._tracking_task_cls = self._build_task_cls(is_tracking_mode=True)
            return self._tracking_task_cls
        else:
            if self._normal_task_cls is None:
                self._normal_task_cls = self._build_task_cls(is_tracking_mode=False)
            return self._normal_task_cls

    def get_task_definition(self):
        return self.get_task_cls().task_definition

    def _build_task_cls(self, is_tracking_mode):
        class_or_func = self.func
        bases = (self.task_type,)
        task_cls = TaskMetaclass(
            str(class_or_func.__name__),
            bases,
            dict(
                _conf__decorator_spec=self.func_spec,
                _callable_item=self._callable_item or class_or_func,
                __doc__=class_or_func.__doc__,
                __module__=class_or_func.__module__,
                defaults=self.task_defaults,
                is_tracking_mode=is_tracking_mode,
            ),
        )
        return task_cls

    def _call_count_limit_exceeded(self):
        if not self._call_as_func:
            self._call_count += 1
            if self._call_count > self._max_call_count:
                logger.info(
                    "Reached maximum tracking limit of {} tasks. Running function regularly.".format(
                        self._max_call_count
                    )
                )
                self._call_as_func = True
        return self._call_as_func

    @dbnd_handle_errors(exit_on_error=False)
    def dbnd_run(self, *args, **kwargs):
        # type: (...)-> DatabandRun
        """
        Run task via Databand execution system
        """
        t = self._build_task(*args, **kwargs)
        return t.dbnd_run()

    @dbnd_handle_errors(exit_on_error=False)
    def _build_task(self, *args, **kwargs):
        task_cls = self.get_task_cls()
        return task_cls(*args, **kwargs)

    @contextlib.contextmanager
    def tracking_context(self, call_args, call_kwargs):
        user_code_called = False  # whether we got to executing of user code
        user_code_finished = False  # whether we passed executing of user code
        func_call = None
        try:
            func_call = FuncCallWithResult(
                task_cls=self.get_task_cls(),
                call_user_code=self.func,
                call_args=tuple(call_args),  # prevent original call_args modification
                call_kwargs=dict(call_kwargs),  # prevent original kwargs modification
            )

            # 1. check that we don't have too many calls
            # 2. Start or reuse existing "inplace_task" that is root for tracked tasks
            if not self._call_count_limit_exceeded() and _get_or_create_inplace_task():
                task_run = _create_dynamic_task_run(func_call)
                with task_run.runner.task_run_execution_context(handle_sigterm=True):
                    task_run.set_task_run_state(state=TaskRunState.RUNNING)

                    _log_inputs(task_run)

                    # if we reached this line, then all tracking initialization is
                    # finished successfully, and we're going to execute user code
                    user_code_called = True

                    try:
                        # tracking_context is context manager - user code will run on yield
                        yield func_call.set_result

                        # if we reached this line, this means that user code finished
                        # successfully without any exceptions
                        user_code_finished = True
                    except Exception as ex:
                        task_run.finished_time = utcnow()

                        error = TaskRunError.build_from_ex(ex, task_run)
                        task_run.set_task_run_state(TaskRunState.FAILED, error=error)
                        raise
                    else:
                        task_run.finished_time = utcnow()

                        # func_call.result should contain result, log it
                        _log_result(task_run, func_call.result)

                        task_run.set_task_run_state(TaskRunState.SUCCESS)
        except Exception:
            if user_code_called and not user_code_finished:
                # if we started to call the user code and not got to user_code_finished
                # line - it means there was user code exception - so just re-raise it
                raise
            # else it's either we didn't reached calling user code, or already passed it
            # then it's some dbnd tracking error - just log it
            if func_call:
                _handle_dynamic_error("tracking-init", func_call)

        # if we didn't reached user_code_called=True line - there was an error during
        # dbnd tracking initialization, so nothing is done - user function wasn't called yet
        if not user_code_called:
            # tracking_context is context manager - user code will run on yield
            yield _passthrough_decorator


class _DecoratedUserClassMeta(type):
    def __call__(cls, *args, **kwargs):
        """
        wrap user class ,so on user_class() we run _item_call first and if required we return task object inplace
        """
        return _call_handler(
            cls.task_cls,
            call_user_code=super(_DecoratedUserClassMeta, cls).__call__,
            call_args=args,
            call_kwargs=kwargs,
        )


def _call_handler(task_cls, call_user_code, call_args, call_kwargs):
    """
    -= Use "Step into My Code"" to get back from Databand code! =-

    decorated object call/creation  ( my_func(), MyDecoratedTask()
    """
    force_invoke = call_kwargs.pop("__force_invoke", False)
    dbnd_project_config = get_dbnd_project_config()

    if force_invoke or dbnd_project_config.disabled:
        # 1. Databand is not enabled
        # 2. we have this call coming from Task.run / Task.band direct invocation
        return call_user_code(*call_args, **call_kwargs)
    func_call = FuncCall(
        task_cls=task_cls,
        call_args=call_args,
        call_kwargs=call_kwargs,
        call_user_code=call_user_code,
    )

    if is_in_airflow_dag_build_context():  # we are in Airflow DAG building mode
        return build_task_at_airflow_dag_context(
            task_cls=task_cls, call_args=call_args, call_kwargs=call_kwargs
        )

    current = try_get_current_task()
    if not current:
        from dbnd._core.inplace_run.inplace_run_manager import try_get_inplace_task_run

        task_run = try_get_inplace_task_run()
        if task_run:
            current = task_run.task

    if not current:  # direct call to the function
        return func_call.invoke()

    ######
    # DBND HANDLING OF CALL
    # now we can make some decisions what we do with the call
    # it's not coming from _invoke_func
    # but from   user code ...   some_func()  or SomeTask()
    phase = current_phase()
    if phase is TaskContextPhase.BUILD:
        # we are in the @pipeline context, we are building execution plan
        t = task_cls(*call_args, **call_kwargs)

        # we are in inline debug mode -> we are going to execute the task
        # we are in the band
        # and want to return result of the object
        if t.task_definition.single_result_output:
            return t.result

        # we have multiple outputs ( result, another output.. )
        # -> just return task object
        return t

    if phase is TaskContextPhase.RUN:
        # we are in the run function!
        if (
            current.settings.dynamic_task.enabled
            and current.task_supports_dynamic_tasks
        ):
            # isinstance() check required to prevent infinite recursion when @task is on
            # class and not on func (example: see test_task_decorated_class.py)
            # and the current task supports inline calls
            # that's extra mechanism in addition to __force_invoke
            # on pickle/unpickle isinstance fails to run.
            return create_and_run_dynamic_task_safe(func_call=func_call)

    # we can not call it in"databand" way, fallback to normal execution
    return func_call.invoke()


def _passthrough_decorator(f):
    return f


def _is_tracking_mode():
    return get_dbnd_project_config().is_tracking_mode()


def _get_or_create_inplace_task():
    """
    try to get existing task, and if not exists - try to get/create inplace_task_run
    """
    current_task = try_get_current_task()
    if not current_task:
        from dbnd._core.inplace_run.inplace_run_manager import try_get_inplace_task_run

        inplace_task_run = try_get_inplace_task_run()
        if inplace_task_run:
            current_task = inplace_task_run.task
    return current_task


def _create_dynamic_task_run(func_call):
    task = _create_dynamic_task(func_call)
    dbnd_run = get_databand_run()
    task_run = dbnd_run.create_dynamic_task_run(
        task, task_engine=current_task_run().task_engine
    )
    return task_run


def _create_dynamic_task(func_call):
    # type: (FuncCall) -> Task
    task_cls = func_call.task_cls
    call_args, call_kwargs = args_to_kwargs(
        task_cls._conf__decorator_spec.args, func_call.call_args, func_call.call_kwargs
    )

    # instantiate inline task
    t = task_cls._create_task(call_args, call_kwargs)

    # update upstream/downstream relations - needed for correct tracking
    # we can have the task as upstream , as it was executed already
    parent_task = current_task_run().task
    if not parent_task.task_dag.has_upstream(t):
        parent_task.set_upstream(t)
    return t


def _log_inputs(task_run):
    """
    For tracking mode. Logs InMemoryTarget inputs.
    """
    try:
        params = task_run.task._params
        for param in params.get_params(input_only=True):
            value = params.get_value(param.name)

            # we
            if isinstance(value, InMemoryTarget):
                try:
                    task_run.tracker.log_parameter_data(
                        parameter=param,
                        target=value,
                        value=value._obj,
                        operation_type=DbndTargetOperationType.read,
                        operation_status=DbndTargetOperationStatus.OK,
                    )
                except Exception as ex:
                    log_exception(
                        "Failed to log input param to tracking store.",
                        ex=ex,
                        non_critical=True,
                    )
    except Exception as ex:
        log_exception(
            "Failed to log input params to tracking store.", ex=ex, non_critical=True
        )


def _log_result(task_run, result):
    """
    For tracking mode. Logs the task result and adds it to the target_origin map to support relationships between
    dynamic tasks.
    """
    try:
        task_result_parameter = task_run.task._params.get_param(RESULT_PARAM)

        # spread result into relevant fields.
        if isinstance(task_result_parameter, FuncResultParameter):
            # assign all returned values to relevant band Outputs
            if result is None:
                return
            for r_name, value in task_result_parameter.named_results(result):
                _log_parameter_value(
                    task_run,
                    parameter_definition=task_run.task._params.get_param(r_name),
                    target=task_run.task._params.get_value(r_name),
                    value=value,
                )
        else:
            _log_parameter_value(
                task_run,
                parameter_definition=task_result_parameter,
                target=task_run.task.result,
                value=result,
            )
    except Exception as ex:
        log_exception(
            "Failed to log result to tracking store.", ex=ex, non_critical=True
        )


def _log_parameter_value(task_run, parameter_definition, target, value):
    try:
        # case what if result is Proxy
        value_type = get_value_type_of_obj(value, parameter_definition.value_type)
        task_run.run.target_origin.add(target, value, value_type)
    except Exception as ex:
        log_exception(
            "Failed to register result to target tracking.", ex=ex, non_critical=True
        )

    try:
        task_run.tracker.log_parameter_data(
            parameter=parameter_definition,  # was: task_run.task.task_definition.task_class.result,
            target=target,
            value=value,
            operation_type=DbndTargetOperationType.write,  # is it write? (or log?)
            operation_status=DbndTargetOperationStatus.OK,
        )
    except Exception as ex:
        log_exception(
            "Failed to log result to tracking store.", ex=ex, non_critical=True
        )


def _task_decorator(*decorator_args, **decorator_kwargs):
    if not is_databand_enabled():
        # simple `@task` decorator, no options were (probably) given.
        if len(decorator_args) == 1 and callable(decorator_args[0]):
            return decorator_args[0]
        return _passthrough_decorator

    task_type = decorator_kwargs.pop("_task_type")  # type: Type[_DecoratedTask]
    task_default_result = decorator_kwargs.pop(
        "_task_default_result"
    )  # ParameterFactory
    task_defaults = decorator_kwargs.pop("defaults", None)

    def decorated(class_or_func):
        try:
            func_spec = build_task_decorator_spec(
                class_or_func=class_or_func,
                decorator_kwargs=decorator_kwargs,
                default_result=task_default_result,
            )
        except Exception as ex:
            logger.error(
                "Failed to create task %s: %s\n%s\n",
                class_or_func.__name__,
                str(ex),
                user_side_code(context=5),
                exc_info=show_exc_info(ex),
            )
            raise

        fp = TaskClsBuilder(func_spec, task_type, task_defaults)

        if func_spec.is_class:
            wrapper = six.add_metaclass(_DecoratedUserClassMeta)(class_or_func)
            fp._callable_item = wrapper
        else:

            @functools.wraps(class_or_func)
            def wrapper(*args, **kwargs):
                if _is_tracking_mode():
                    with fp.tracking_context(args, kwargs) as track_result_callback:
                        return track_result_callback(fp.func(*args, **kwargs))

                return _call_handler(
                    fp.get_task_cls(),
                    call_user_code=fp.func,
                    call_args=args,
                    call_kwargs=kwargs,
                )

            wrapper.dbnd_run = fp.dbnd_run

        wrapper.__is_dbnd_task__ = True
        wrapper.func = class_or_func

        # we're using CallableLazyObjectProxy to have lazy evaluation for creating task_cls
        task_cls = CallableLazyObjectProxy(fp.get_task_cls)
        wrapper.task_cls = task_cls
        wrapper.task = task_cls
        wrapper.t = task_cls

        # we need lazy task_definition here, for example for dbnd_task_as_bash_operator
        wrapper.task_definition = CallableLazyObjectProxy(fp.get_task_definition)

        # we need to manually register the task here, since in regular flow
        # this happens in TaskMetaclass, but it's not invoked here due to lazy
        # evaluation using CallableLazyObjectProxy
        tp = TaskPassport.from_func_spec(func_spec, decorator_kwargs)

        # TODO: we can use CallableLazyObjectProxy object (task_cls) instead of task_cls_factory
        r = get_task_registry()
        r.register_task_cls_factory(
            task_cls_factory=fp.get_task_cls,
            full_task_family=tp.full_task_family,
            task_family=tp.task_family,
        )

        return wrapper

    # simple `@task` decorator, no options were (probably) given.
    if len(decorator_args) == 1 and callable(decorator_args[0]):
        return decorated(decorator_args[0])

    return decorated



# -*- coding: utf-8 -*-
"""
click._termui_impl
~~~~~~~~~~~~~~~~~~

This module contains implementations for the termui module. To keep the
import time of Click down, some infrequently used functionality is
placed in this module and only imported as needed.

:copyright: © 2014 by the Pallets team.
:license: BSD, see LICENSE.rst for more details.
"""

import os
import sys
import time
import math
import contextlib
from ._compat import _default_text_stdout, range_type, isatty, \
     open_stream, strip_ansi, term_len, get_best_encoding, WIN, int_types, \
     CYGWIN
from .utils import echo
from .exceptions import ClickException


if os.name == 'nt':
    BEFORE_BAR = '\r'
    AFTER_BAR = '\n'
else:
    BEFORE_BAR = '\r\033[?25l'
    AFTER_BAR = '\033[?25h\n'


def _length_hint(obj):
    """Returns the length hint of an object."""
    try:
        return len(obj)
    except (AttributeError, TypeError):
        try:
            get_hint = type(obj).__length_hint__
        except AttributeError:
            return None
        try:
            hint = get_hint(obj)
        except TypeError:
            return None
        if hint is NotImplemented or \
           not isinstance(hint, int_types) or \
           hint < 0:
            return None
        return hint


class ProgressBar(object):

    def __init__(self, iterable, length=None, fill_char='#', empty_char=' ',
                 bar_template='%(bar)s', info_sep='  ', show_eta=True,
                 show_percent=None, show_pos=False, item_show_func=None,
                 label=None, file=None, color=None, width=30):
        self.fill_char = fill_char
        self.empty_char = empty_char
        self.bar_template = bar_template
        self.info_sep = info_sep
        self.show_eta = show_eta
        self.show_percent = show_percent
        self.show_pos = show_pos
        self.item_show_func = item_show_func
        self.label = label or ''
        if file is None:
            file = _default_text_stdout()
        self.file = file
        self.color = color
        self.width = width
        self.autowidth = width == 0

        if length is None:
            length = _length_hint(iterable)
        if iterable is None:
            if length is None:
                raise TypeError('iterable or length is required')
            iterable = range_type(length)
        self.iter = iter(iterable)
        self.length = length
        self.length_known = length is not None
        self.pos = 0
        self.avg = []
        self.start = self.last_eta = time.time()
        self.eta_known = False
        self.finished = False
        self.max_width = None
        self.entered = False
        self.current_item = None
        self.is_hidden = not isatty(self.file)
        self._last_line = None
        self.short_limit = 0.5

    def __enter__(self):
        self.entered = True
        self.render_progress()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.render_finish()

    def __iter__(self):
        if not self.entered:
            raise RuntimeError('You need to use progress bars in a with block.')
        self.render_progress()
        return self.generator()

    def is_fast(self):
        return time.time() - self.start <= self.short_limit

    def render_finish(self):
        if self.is_hidden or self.is_fast():
            return
        self.file.write(AFTER_BAR)
        self.file.flush()

    @property
    def pct(self):
        if self.finished:
            return 1.0
        return min(self.pos / (float(self.length) or 1), 1.0)

    @property
    def time_per_iteration(self):
        if not self.avg:
            return 0.0
        return sum(self.avg) / float(len(self.avg))

    @property
    def eta(self):
        if self.length_known and not self.finished:
            return self.time_per_iteration * (self.length - self.pos)
        return 0.0

    def format_eta(self):
        if self.eta_known:
            t = int(self.eta)
            seconds = t % 60
            t //= 60
            minutes = t % 60
            t //= 60
            hours = t % 24
            t //= 24
            if t > 0:
                days = t
                return '%dd %02d:%02d:%02d' % (days, hours, minutes, seconds)
            else:
                return '%02d:%02d:%02d' % (hours, minutes, seconds)
        return ''

    def format_pos(self):
        pos = str(self.pos)
        if self.length_known:
            pos += '/%s' % self.length
        return pos

    def format_pct(self):
        return ('% 4d%%' % int(self.pct * 100))[1:]

    def format_bar(self):
        if self.length_known:
            bar_length = int(self.pct * self.width)
            bar = self.fill_char * bar_length
            bar += self.empty_char * (self.width - bar_length)
        elif self.finished:
            bar = self.fill_char * self.width
        else:
            bar = list(self.empty_char * (self.width or 1))
            if self.time_per_iteration != 0:
                bar[int((math.cos(self.pos * self.time_per_iteration)
                    / 2.0 + 0.5) * self.width)] = self.fill_char
            bar = ''.join(bar)
        return bar

    def format_progress_line(self):
        show_percent = self.show_percent

        info_bits = []
        if self.length_known and show_percent is None:
            show_percent = not self.show_pos

        if self.show_pos:
            info_bits.append(self.format_pos())
        if show_percent:
            info_bits.append(self.format_pct())
        if self.show_eta and self.eta_known and not self.finished:
            info_bits.append(self.format_eta())
        if self.item_show_func is not None:
            item_info = self.item_show_func(self.current_item)
            if item_info is not None:
                info_bits.append(item_info)

        return (self.bar_template % {
            'label': self.label,
            'bar': self.format_bar(),
            'info': self.info_sep.join(info_bits)
        }).rstrip()

    def render_progress(self):
        from .termui import get_terminal_size

        if self.is_hidden:
            return

        buf = []
        # Update width in case the terminal has been resized
        if self.autowidth:
            old_width = self.width
            self.width = 0
            clutter_length = term_len(self.format_progress_line())
            new_width = max(0, get_terminal_size()[0] - clutter_length)
            if new_width < old_width:
                buf.append(BEFORE_BAR)
                buf.append(' ' * self.max_width)
                self.max_width = new_width
            self.width = new_width

        clear_width = self.width
        if self.max_width is not None:
            clear_width = self.max_width

        buf.append(BEFORE_BAR)
        line = self.format_progress_line()
        line_len = term_len(line)
        if self.max_width is None or self.max_width < line_len:
            self.max_width = line_len

        buf.append(line)
        buf.append(' ' * (clear_width - line_len))
        line = ''.join(buf)
        # Render the line only if it changed.

        if line != self._last_line and not self.is_fast():
            self._last_line = line
            echo(line, file=self.file, color=self.color, nl=False)
            self.file.flush()

    def make_step(self, n_steps):
        self.pos += n_steps
        if self.length_known and self.pos >= self.length:
            self.finished = True

        if (time.time() - self.last_eta) < 1.0:
            return

        self.last_eta = time.time()

        # self.avg is a rolling list of length <= 7 of steps where steps are
        # defined as time elapsed divided by the total progress through
        # self.length.
        if self.pos:
            step = (time.time() - self.start) / self.pos
        else:
            step = time.time() - self.start

        self.avg = self.avg[-6:] + [step]

        self.eta_known = self.length_known

    def update(self, n_steps, current_item=None):
        """Update the progress bar by advancing a specified number of
        steps, and optionally set the ``current_item`` for this new
        position.

        :param n_steps: Number of steps to advance.
        :param current_item: Optional item to set as ``current_item``
            for the updated position.

        .. versionadded:: 8.0
            Added the ``current_item`` optional parameter.
        """
        self.make_step(n_steps)
        if current_item is not None:
            self.current_item = current_item
        self.render_progress()

    def finish(self):
        self.eta_known = 0
        self.current_item = None
        self.finished = True

    def generator(self):
        """
        Returns a generator which yields the items added to the bar during
        construction, and updates the progress bar *after* the yielded block
        returns.
        """
        if not self.entered:
            raise RuntimeError('You need to use progress bars in a with block.')

        if self.is_hidden:
            for rv in self.iter:
                yield rv
        else:
            for rv in self.iter:
                self.current_item = rv
                yield rv
                self.update(1)
            self.finish()
            self.render_progress()


def pager(generator, color=None):
    """Decide what method to use for paging through text."""
    stdout = _default_text_stdout()
    if not isatty(sys.stdin) or not isatty(stdout):
        return _nullpager(stdout, generator, color)
    pager_cmd = (os.environ.get('PAGER', None) or '').strip()
    if pager_cmd:
        if WIN:
            return _tempfilepager(generator, pager_cmd, color)
        return _pipepager(generator, pager_cmd, color)
    if os.environ.get('TERM') in ('dumb', 'emacs'):
        return _nullpager(stdout, generator, color)
    if WIN or sys.platform.startswith('os2'):
        return _tempfilepager(generator, 'more <', color)
    if hasattr(os, 'system') and os.system('(less) 2>/dev/null') == 0:
        return _pipepager(generator, 'less', color)

    import tempfile
    fd, filename = tempfile.mkstemp()
    os.close(fd)
    try:
        if hasattr(os, 'system') and os.system('more "%s"' % filename) == 0:
            return _pipepager(generator, 'more', color)
        return _nullpager(stdout, generator, color)
    finally:
        os.unlink(filename)


def _pipepager(generator, cmd, color):
    """Page through text by feeding it to another program.  Invoking a
    pager through this might support colors.
    """
    import subprocess
    env = dict(os.environ)

    # If we're piping to less we might support colors under the
    # condition that
    cmd_detail = cmd.rsplit('/', 1)[-1].split()
    if color is None and cmd_detail[0] == 'less':
        less_flags = os.environ.get('LESS', '') + ' '.join(cmd_detail[1:])
        if not less_flags:
            env['LESS'] = '-R'
            color = True
        elif 'r' in less_flags or 'R' in less_flags:
            color = True

    c = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                         env=env)
    encoding = get_best_encoding(c.stdin)
    try:
        for text in generator:
            if not color:
                text = strip_ansi(text)

            c.stdin.write(text.encode(encoding, 'replace'))
    except (IOError, KeyboardInterrupt):
        pass
    else:
        c.stdin.close()

    # Less doesn't respect ^C, but catches it for its own UI purposes (aborting
    # search or other commands inside less).
    #
    # That means when the user hits ^C, the parent process (click) terminates,
    # but less is still alive, paging the output and messing up the terminal.
    #
    # If the user wants to make the pager exit on ^C, they should set
    # `LESS='-K'`. It's not our decision to make.
    while True:
        try:
            c.wait()
        except KeyboardInterrupt:
            pass
        else:
            break


def _tempfilepager(generator, cmd, color):
    """Page through text by invoking a program on a temporary file."""
    import tempfile
    filename = tempfile.mktemp()
    # TODO: This never terminates if the passed generator never terminates.
    text = "".join(generator)
    if not color:
        text = strip_ansi(text)
    encoding = get_best_encoding(sys.stdout)
    with open_stream(filename, 'wb')[0] as f:
        f.write(text.encode(encoding))
    try:
        os.system(cmd + ' "' + filename + '"')
    finally:
        os.unlink(filename)


def _nullpager(stream, generator, color):
    """Simply print unformatted text.  This is the ultimate fallback."""
    for text in generator:
        if not color:
            text = strip_ansi(text)
        stream.write(text)


class Editor(object):

    def __init__(self, editor=None, env=None, require_save=True,
                 extension='.txt'):
        self.editor = editor
        self.env = env
        self.require_save = require_save
        self.extension = extension

    def get_editor(self):
        if self.editor is not None:
            return self.editor
        for key in 'VISUAL', 'EDITOR':
            rv = os.environ.get(key)
            if rv:
                return rv
        if WIN:
            return 'notepad'
        for editor in 'vim', 'nano':
            if os.system('which %s >/dev/null 2>&1' % editor) == 0:
                return editor
        return 'vi'

    def edit_file(self, filename):
        import subprocess
        editor = self.get_editor()
        if self.env:
            environ = os.environ.copy()
            environ.update(self.env)
        else:
            environ = None
        try:
            c = subprocess.Popen('%s "%s"' % (editor, filename),
                                 env=environ, shell=True)
            exit_code = c.wait()
            if exit_code != 0:
                raise ClickException('%s: Editing failed!' % editor)
        except OSError as e:
            raise ClickException('%s: Editing failed: %s' % (editor, e))

    def edit(self, text):
        import tempfile

        text = text or ''
        binary_data = type(text) in [bytes, bytearray]

        if not binary_data and text and not text.endswith('\n'):
            text += '\n'

        fd, name = tempfile.mkstemp(prefix='editor-', suffix=self.extension)
        try:
            if not binary_data:
                if WIN:
                    encoding = 'utf-8-sig'
                    text = text.replace('\n', '\r\n')
                else:
                    encoding = 'utf-8'
                text = text.encode(encoding)

            f = os.fdopen(fd, 'wb')
            f.write(text)
            f.close()
            timestamp = os.path.getmtime(name)

            self.edit_file(name)

            if self.require_save \
               and os.path.getmtime(name) == timestamp:
                return None

            f = open(name, 'rb')
            try:
                rv = f.read()
            finally:
                f.close()
            if binary_data:
                return rv
            else:
                return rv.decode('utf-8-sig').replace('\r\n', '\n')
        finally:
            os.unlink(name)


def open_url(url, wait=False, locate=False):
    import subprocess

    def _unquote_file(url):
        try:
            import urllib
        except ImportError:
            import urllib
        if url.startswith('file://'):
            url = urllib.unquote(url[7:])
        return url

    if sys.platform == 'darwin':
        args = ['open']
        if wait:
            args.append('-W')
        if locate:
            args.append('-R')
        args.append(_unquote_file(url))
        null = open('/dev/null', 'w')
        try:
            return subprocess.Popen(args, stderr=null).wait()
        finally:
            null.close()
    elif WIN:
        if locate:
            url = _unquote_file(url)
            args = 'explorer /select,"%s"' % _unquote_file(
                url.replace('"', ''))
        else:
            args = 'start %s "" "%s"' % (
                wait and '/WAIT' or '', url.replace('"', ''))
        return os.system(args)
    elif CYGWIN:
        if locate:
            url = _unquote_file(url)
            args = 'cygstart "%s"' % (os.path.dirname(url).replace('"', ''))
        else:
            args = 'cygstart %s "%s"' % (
                wait and '-w' or '', url.replace('"', ''))
        return os.system(args)

    try:
        if locate:
            url = os.path.dirname(_unquote_file(url)) or '.'
        else:
            url = _unquote_file(url)
        c = subprocess.Popen(['xdg-open', url])
        if wait:
            return c.wait()
        return 0
    except OSError:
        if url.startswith(('http://', 'https://')) and not locate and not wait:
            import webbrowser
            webbrowser.open(url)
            return 0
        return 1


def _translate_ch_to_exc(ch):
    if ch == u'\x03':
        raise KeyboardInterrupt()
    if ch == u'\x04' and not WIN:  # Unix-like, Ctrl+D
        raise EOFError()
    if ch == u'\x1a' and WIN:      # Windows, Ctrl+Z
        raise EOFError()


if WIN:
    import msvcrt

    @contextlib.contextmanager
    def raw_terminal():
        yield

    def getchar(echo):
        # The function `getch` will return a bytes object corresponding to
        # the pressed character. Since Windows 10 build 1803, it will also
        # return \x00 when called a second time after pressing a regular key.
        #
        # `getwch` does not share this probably-bugged behavior. Moreover, it
        # returns a Unicode object by default, which is what we want.
        #
        # Either of these functions will return \x00 or \xe0 to indicate
        # a special key, and you need to call the same function again to get
        # the "rest" of the code. The fun part is that \u00e0 is
        # "latin small letter a with grave", so if you type that on a French
        # keyboard, you _also_ get a \xe0.
        # E.g., consider the Up arrow. This returns \xe0 and then \x48. The
        # resulting Unicode string reads as "a with grave" + "capital H".
        # This is indistinguishable from when the user actually types
        # "a with grave" and then "capital H".
        #
        # When \xe0 is returned, we assume it's part of a special-key sequence
        # and call `getwch` again, but that means that when the user types
        # the \u00e0 character, `getchar` doesn't return until a second
        # character is typed.
        # The alternative is returning immediately, but that would mess up
        # cross-platform handling of arrow keys and others that start with
        # \xe0. Another option is using `getch`, but then we can't reliably
        # read non-ASCII characters, because return values of `getch` are
        # limited to the current 8-bit codepage.
        #
        # Anyway, Click doesn't claim to do this Right(tm), and using `getwch`
        # is doing the right thing in more situations than with `getch`.
        if echo:
            func = msvcrt.getwche
        else:
            func = msvcrt.getwch

        rv = func()
        if rv in (u'\x00', u'\xe0'):
            # \x00 and \xe0 are control characters that indicate special key,
            # see above.
            rv += func()
        _translate_ch_to_exc(rv)
        return rv
else:
    import tty
    import termios

    @contextlib.contextmanager
    def raw_terminal():
        if not isatty(sys.stdin):
            f = open('/dev/tty')
            fd = f.fileno()
        else:
            fd = sys.stdin.fileno()
            f = None
        try:
            old_settings = termios.tcgetattr(fd)
            try:
                tty.setraw(fd)
                yield fd
            finally:
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                sys.stdout.flush()
                if f is not None:
                    f.close()
        except termios.error:
            pass

    def getchar(echo):
        with raw_terminal() as fd:
            ch = os.read(fd, 32)
            ch = ch.decode(get_best_encoding(sys.stdin), 'replace')
            if echo and isatty(sys.stdout):
                sys.stdout.write(ch)
            _translate_ch_to_exc(ch)
            return ch



from typing import Union

import attr

from jinja2 import Template


@attr.s
class StringRenderer(object):
    template = attr.ib()  # type: Union[str,Template]
    jinja_mode = attr.ib(default=False)  # type: bool

    @classmethod
    def from_str(cls, template):
        jinja_mode = "{{" in template
        if jinja_mode:
            template = Template(template)
        return cls(template=template, jinja_mode=jinja_mode)

    def render_str(self, **ctx):
        if self.jinja_mode:
            return self.template.render(**ctx)

        return self.template.format(**ctx)



import enum
import logging
import typing

from typing import Any, Dict, Optional, Type

import attr
import six

from dbnd._core.configuration.config_path import ConfigPath
from dbnd._core.configuration.config_value import ConfigValue
from dbnd._core.configuration.environ_config import PARAM_ENV_TEMPLATE
from dbnd._core.constants import (
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    OutputMode,
    _TaskParamContainer,
)
from dbnd._core.current import try_get_databand_run
from dbnd._core.errors import DatabandBuildError, friendly_error
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.parameter.parameter_value import ParameterValue
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._core.utils.basics.text_banner import safe_string
from dbnd._core.utils.traversing import traverse
from dbnd._vendor.snippets.airflow_configuration import expand_env_var
from targets.base_target import Target, TargetSource
from targets.data_target import DataTarget
from targets.file_target import FileTarget
from targets.inline_target import InlineTarget
from targets.inmemory_target import InMemoryTarget
from targets.target_config import FileFormat, TargetConfig
from targets.target_factory import target
from targets.types import Path
from targets.value_meta import ValueMetaConf
from targets.values import (
    InlineValueType,
    StrValueType,
    ValueType,
    VersionValueType,
    get_types_registry,
)
from targets.values.builtins_values import DefaultObjectValueType
from targets.values.structure import _StructureValueType
from targets.values.target_values import _TargetValueType


logger = logging.getLogger(__name__)
if typing.TYPE_CHECKING:
    from dbnd._core.task.task import Task


class ParameterScope(enum.Enum):
    task = "task"  # do not propagate value
    children = "children"  # propagate value to direct children


class _ParameterKind(enum.Enum):
    task_input = "task_input"
    task_output = "task_output"


class ParameterGroup(enum.Enum):
    system = "system"
    user = "user"


T = typing.TypeVar("T")


@attr.s(repr=False, hash=False, str=False)
class ParameterDefinition(object):  # generics are broken: typing.Generic[T]
    """
    Parameter whose value is a ``str``, and a base class for other parameter types.

    Parameters are objects set on the Task class level to make it possible to parameterize tasks.
    For instance:

    .. code:: python

        class MyTask(dbnd.Task):
            foo = databand.parameter[str]

        class RequiringTask(dbnd.Task):
            def requires(self):
                return MyTask(foo="hello")

            def run(self):
                print(self.requires().foo)  # prints "hello"

    This makes it possible to instantiate multiple tasks, eg ``MyTask(foo='bar')`` and
    ``MyTask(foo='baz')``. The task will then have the ``foo`` attribute set appropriately.

    When a task is instantiated, it will first use any argument as the value of the parameter, eg.
    if you instantiate ``a = TaskA(x=44)`` then ``a.x == 44``. When the value is not provided, the
    value  will be resolved in this order of falling priority:

        * Any value provided on the command line:

          - To the root task (eg. ``--param xyz``)

          - Then to the class, using the qualified task name syntax (eg. ``--TaskA-param xyz``).

        * With ``[TASK_NAME]>PARAM_NAME: <serialized value>`` syntax. See :ref:`ParamConfigIngestion`

        * Any default value set using the ``default`` flag.

    """

    default_description = ""
    default_input_description = "data input"
    default_output_description = "data output"

    _total_counter = (
        0  # non-atomically increasing counter used for ordering parameters.
    )

    default = attr.ib(default=NOTHING)

    name = attr.ib(default=None)

    # value type and sub type
    value_type = attr.ib(default=None)  # type: ValueType
    value_type_defined = attr.ib(default=None)  # type: ValueType
    sub_type = attr.ib(default=None)

    description = attr.ib(default=NOTHING)  # type: str
    config_path = attr.ib(default=None)  # type: Optional[ConfigPath]
    disable_jinja_templating = attr.ib(default=False)  # type: bool
    require_local_access = attr.ib(default=False)  # type: bool
    env_interpolation = attr.ib(default=True)
    # parameter kind
    significant = attr.ib(default=True)  # type: bool
    scope = attr.ib(default=ParameterScope.task)
    system = attr.ib(default=False)
    kind = attr.ib(default=_ParameterKind.task_input)

    # output configuration
    output_name = attr.ib(default=None)  # type: str
    output_ext = attr.ib(default=None)  # type: str
    output_mode = attr.ib(default=OutputMode.regular)
    # used for output factories only
    output_factory = attr.ib(default=None)

    target_config = attr.ib(default=TargetConfig())
    load_options = attr.ib(factory=dict)  # type: Dict[FileFormat, Dict[str, Any]]
    save_options = attr.ib(factory=dict)  # type: Dict[FileFormat, Dict[str, Any]]

    validator = attr.ib(default=None)
    choices = attr.ib(default=None)

    load_on_build = attr.ib(default=NOTHING)  # type: bool
    empty_default = attr.ib(default=NOTHING)

    # value preview and meta settings
    log_preview = attr.ib(default=None)  # type: Optional[bool]
    log_preview_size = attr.ib(default=None)  # type: Optional[int]
    log_schema = attr.ib(default=None)  # type: Optional[bool]
    log_size = attr.ib(default=None)  # type: Optional[bool]
    log_stats = attr.ib(default=None)  # type: Optional[bool]
    log_histograms = attr.ib(default=None)  # type: Optional[bool]

    log_meta = attr.ib(
        default=True
    )  # type: bool  # log all (can disable whole value log)

    # ParameterDefinition ownership
    task_cls = attr.ib(default=None)  # type: Type[Task]
    parameter_origin = attr.ib(default=None)
    parameter_id = attr.ib(default=1)

    value_meta_conf = attr.ib(default=None)  # type: ValueMetaConf
    hidden = attr.ib(default=False)  # type: bool

    @property
    def group(self):
        return ParameterGroup.system if self.system else ParameterGroup.user

    @property
    def task_definition_uid(self):
        if not self.task_cls:
            return None
        return self.task_cls.task_definition.task_definition_uid

    def evolve_with_owner(self, task_cls, name):
        if self.task_cls and self.name != name:
            logger.warning(
                "Name of parameter has been changed from '%s' to '%s' at %s",
                name,
                self.name,
                task_cls,
            )
        parameter_origin = self.parameter_origin or task_cls
        return attr.evolve(
            self, task_cls=task_cls, name=name, parameter_origin=parameter_origin
        )

    def parse_from_str(self, x):  # type: (str) -> T
        """
        Parse an individual value from the input.

        :param str x: the value to parse.
        :return: the parsed value.
        """
        return self.calc_init_value(x)

    def calc_init_value(self, value):
        if value is None:
            # it's None
            # TODO: may be we still can "normalize" the value
            return value
        if isinstance(value, Path):
            return target(str(value), config=self.target_config)
        if isinstance(value, Target):
            # it's deferred result - > we load it lately
            return value

        # we process value regardless parse!
        # cf_value.require_parse:

        if self.env_interpolation and isinstance(value, six.string_types):
            try:
                value = expand_env_var(value)
            except Exception as ex:
                logger.warning(
                    "failed to expand variable '%s' : %s", safe_string(value), str(ex)
                )

        # in case we are output and have value:
        # it's Target or it's str to be converted as target
        load_value = self.load_on_build and not self.is_output()

        return self.value_type.parse_value(
            value, load_value=load_value, target_config=self.target_config
        )

    def calc_runtime_value(self, value, task):
        if value is None:
            return value

        if isinstance(self.value_type, _TargetValueType):
            # if it "target" type, let read it into "user friendly" format
            # regardless it's input or output

            return traverse(value, self.value_type.target_to_value)

        # usually we should not load "outputs" on read
        if self.is_output():
            # actually we should not load it, so just return
            return value

        if isinstance(value, Target):
            try:
                runtime_value = self.load_from_target(value)
                if self.is_input():
                    self._log_parameter_value(runtime_value, value, task)
                return runtime_value
            except Exception as ex:
                raise friendly_error.failed_to_read_target_as_task_input(
                    ex=ex, task=task, parameter=self, target=value
                )

        if (
            isinstance(self.value_type, _StructureValueType)
            and self.value_type.sub_value_type
        ):
            try:

                def load_with_preview(val):
                    runtime_val = self.value_type.sub_value_type.load_runtime(val)
                    if self.is_input() and isinstance(val, Target):
                        # Optimisation opportunity: log all targets in a single call
                        self._log_parameter_value(runtime_val, val, task)

                    return runtime_val

                return traverse(value, convert_f=load_with_preview)
            except Exception as ex:
                raise friendly_error.failed_to_read_task_input(
                    ex=ex, task=task, parameter=self, target=value
                )

        return value

    def to_str(self, x):  # type: (T) -> str
        """
        Opposite of :py:meth:`parse`.

        Converts the value ``x`` to a string.

        :param x: the value to serialize.
        """
        if isinstance(x, Target):
            return str(x)

        return self.value_type.to_str(x)  # default impl

    def to_repr(self, x):  # type: (T) -> str
        return self.value_type.to_repr(x)

    def signature(self, x):
        if x is None:
            return str(x)

        if isinstance(x, Target):
            return str(x)

        # we can have
        # 1. a value of value_type
        # 2. target with value type TargetValueType
        # 3. list/dict of targets with value type TargetValueType
        return self.value_type.to_signature(x)

    def load_from_target(
        self, target, **kwargs
    ):  # type: (ParameterDefinition, FileTarget, **Any)-> T
        from targets import InMemoryTarget

        if isinstance(target, InMemoryTarget):
            value = target.load()
        else:
            if target.config:
                f = target.config.format
                if f and f in self.load_options:
                    kwargs.update(**self.load_options[f])
            value = self.value_type.load_from_target(target, **kwargs)

        self.validate(value)

        self._store_value_origin_target(value, target)

        return value

    def dump_to_target(
        self, target, value, **kwargs
    ):  # type: (DataTarget, T, **Any)-> None
        if hasattr(target, "config"):
            f = target.config.format
            if f and f in self.save_options:
                kwargs.update(**self.save_options[f])

        self.value_type.save_to_target(target, value, **kwargs)  # default impl

        # we need updated target
        self._store_value_origin_target(value, target)

    def _log_parameter_value(self, runtime_value, value, task):
        if try_get_databand_run() and task.current_task_run:
            task.current_task_run.tracker.log_parameter_data(
                parameter=self,
                target=value,
                value=runtime_value,
                operation_type=DbndTargetOperationType.read,
                operation_status=DbndTargetOperationStatus.OK,
            )

    def _store_value_origin_target(self, value, target):
        dbnd_run = try_get_databand_run()
        if not dbnd_run:
            return

        dbnd_run.target_origin.add(target, value, self.value_type)

    def normalize(self, x):  # type: (T) -> T
        """
        Given a parsed parameter value, normalizes it.

        The value can either be the result of parse(), the default value or
        arguments passed into the task's constructor by instantiation.

        This is very implementation defined, but can be used to validate/clamp
        valid values. For example, if you wanted to only accept even integers,
        and "correct" odd values to the nearest integer, you can implement
        normalize as ``x // 2 * 2``.
        """
        if isinstance(self.value_type, _TargetValueType):
            # can not move to value_type, we need target_config

            from dbnd._core.utils.task_utils import to_targets

            return to_targets(x, from_string_kwargs=dict(config=self.target_config))
        return self.value_type.normalize(x)

    def validate(self, x):
        if self.validator:
            self.validator.validate(self, x)

    def as_str_input(self, value):
        switch_value = self.to_str(value)
        if isinstance(value, Target) and self.load_on_build:
            switch_value = "@%s" % switch_value
        return switch_value

    def next_in_enumeration(self, value):
        """
        If your Parameter type has an enumerable ordering of values. You can
        choose to override this method. This method is used by the
        :py:mod:`databand.execution_summary` module for pretty printing
        purposes. Enabling it to pretty print tasks like ``MyTask(num=1),
        MyTask(num=2), MyTask(num=3)`` to ``MyTask(num=1..3)``.

        :param value: The value
        :return: The next value, like "value + 1". Or ``None`` if there's no enumerable ordering.
        """
        return self.value_type.next_in_enumeration(value)  # default impl

    def _get_help_message(self, sections=None):
        sections = sections or [(self.task_family)]

        define_via = []

        define_via.append(
            "project.cfg : [%s]%s=VALUE" % (" | ".join(sections), self.name)
        )
        define_via.append("cli:   --set %s.%s=VALUE" % (self.task_family, self.name))
        define_via.append(
            "constructor: %s(%s=VALUE, ...)" % (self.task_family, self.name)
        )
        define_via = "\n".join(["\t* %s" % l for l in define_via])

        return "You can change '{task_family}.{name}' value using one of the following methods: \n {methods}".format(
            task_family=(self.task_family), name=self.name, methods=define_via
        )

    def parameter_exception(self, reason, ex):
        err_msg = "Failed to {reason} for parameter '{name}' at {task_family}()".format(
            reason=reason, name=self.name, task_family=self.task_family
        )
        log_exception(err_msg, ex, logger)
        raise DatabandBuildError(
            err_msg, nested_exceptions=[ex], help_msg=self._get_help_message()
        )

    @property
    def task_family(self):
        if self.task_cls:
            return self.task_cls.task_definition.task_family
        return None

    @property
    def task_config_section(self):
        if self.task_cls:
            return self.task_cls.task_definition.task_config_section
        return None

    def __repr__(self):
        owned_by = ""
        parameter_origin = ""  # show it only if different

        if self.task_cls:
            owned_by = self.task_cls.get_task_family() if self.task_cls else ""
            origin_cls_str = (
                self.parameter_origin.get_task_family() if self.parameter_origin else ""
            )
            if origin_cls_str and origin_cls_str != owned_by:
                parameter_origin = " at %s" % origin_cls_str

        parameter_kind = (
            "output" if self.kind == _ParameterKind.task_output else "parameter"
        )
        return "{owned_by}.{name}({parameter_kind}[{value_type}]{parameter_origin})".format(
            owned_by=owned_by,
            parameter_origin=parameter_origin,
            value_type=self.value_type_str,
            parameter_kind=parameter_kind,
            name=self.name or "_unknown_",
        )

    @property
    def value_type_str(self):
        if self.value_type is None:
            return "unknown"
        type_handler = self.value_type.type_str
        if isinstance(self.value_type, InlineValueType):
            type_handler = "!" + type_handler
        if self.value_type_defined != self.value_type:
            type_handler = "*" + type_handler
        return type_handler

    def _target_source(self, task):
        return TargetSource(
            task_id=task.task_id, parameter_name=self.name, name=self.name
        )

    def build_target(self, task):  # type: (ParameterDefinition, Task) -> DataTarget
        target_config = self.target_config
        if not target_config.format:
            default_config = task.settings.output.get_config(self.value_type)
            # for now we take only format and compression from config
            target_config = target_config.with_format(
                default_config.format
            ).with_compression(default_config.compression)
        output_ext = self.output_ext
        if output_ext is None:
            output_ext = target_config.get_ext()
        return task.get_target(
            name=self.output_name or self.name,
            output_ext=output_ext,
            config=target_config,
            output_mode=self.output_mode,
        )

    def build_output(self, task):
        if self.output_factory is not None:
            try:
                return self.output_factory(task, self)
            except Exception:
                logger.exception(
                    "Failed to created task output %s for %s : "
                    " output_factory expected signature is '(Task, Parameter) -> Target(any structure) '",
                    self,
                    task,
                )
                raise

        if (
            not self.system
            and self.name not in ("task_band",)
            and task.task_in_memory_outputs
        ):
            return InMemoryTarget(
                path="memory://{value_type}:{task}.{p_name}".format(
                    value_type=self.value_type, task=task.task_id, p_name=self.name
                )
            )

        # we separate into two functions ,
        # as we want to be able to call build_target from output_factory implementation
        try:
            return self.build_target(task)
        except Exception as e:
            raise friendly_error.task_build.failed_to_build_output_target(
                self.name, task, e
            )

    def is_input(self):
        return self.kind == _ParameterKind.task_input

    def is_output(self):
        return self.kind == _ParameterKind.task_output

    def __hash__(self):
        return hash(self.name) ^ hash(self.task_cls)

    def modify(self, **kwargs):
        if not kwargs:
            return self
        return attr.evolve(self, **kwargs)

    def get_env_key(self, section):
        return PARAM_ENV_TEMPLATE.format(S=section.upper(), K=self.name.upper())

    def get_value_meta(self, value, meta_conf):
        # do not use meta_conf directly, you should get it merged with main config
        return self.value_type.get_value_meta(value, meta_conf=meta_conf)


def _update_parameter_from_runtime_value_type(parameter, value):
    # type: ( ParameterDefinition, Any)->Optional[ValueType]

    original_value_type = parameter.value_type
    runtime_value_type = None
    if isinstance(value, Target):
        if isinstance(value, InlineTarget):
            runtime_value_type = value.value_type
        elif isinstance(original_value_type, _TargetValueType):
            # user expects to get target/path/str
            # we will validate that value is good at "parse"
            pass
        else:
            # we are going to "load" the value from target into original_value_type
            # let get the "real" value type from the source of it
            if value.source and value.source_parameter:
                # we can take value type from the creator of it
                runtime_value_type = value.source_parameter.value_type
                if isinstance(
                    runtime_value_type, (_TargetValueType, DefaultObjectValueType)
                ):
                    return None
            else:
                # we don't really have value type
                return None
    else:
        if isinstance(value, six.string_types) or isinstance(value, Path):
            # str are going to be parsed, or treated as Target
            # Path is going to be used as Target
            return None

        runtime_value_type = get_types_registry().get_value_type_of_obj(value)

    # not found or the same
    if not runtime_value_type or runtime_value_type == original_value_type:
        return None

    # value is str, most chances we will parse it into the value
    if type(runtime_value_type) in [StrValueType, VersionValueType]:
        return None
    if isinstance(runtime_value_type, _StructureValueType):
        if isinstance(original_value_type, _StructureValueType):
            if type(runtime_value_type) == type(original_value_type):
                # probably we have difference on sub level,
                # there is no a clear way to find sub types of object
                if not runtime_value_type.sub_value_type:
                    return None
    if original_value_type.type is object:
        if runtime_value_type.type is object:
            return None
        if not isinstance(original_value_type, DefaultObjectValueType):
            return None
    return runtime_value_type


def build_parameter_value(parameter, cf_value):
    # type: (ParameterDefinition, ConfigValue) -> ParameterValue
    if not cf_value:
        return ParameterValue(
            parameter=parameter,
            source=None,
            source_value=NOTHING,
            value=NOTHING,
            parsed=False,
        )

    warnings = []
    value = cf_value.value
    try:
        if value is not None and not parameter.is_output():
            updated_value_type = _update_parameter_from_runtime_value_type(
                parameter, value
            )
            message = (
                "{parameter}: type of the value at runtime '{runtime}'"
                " doesn't match user defined type '{compile}'".format(
                    parameter=parameter,
                    runtime=updated_value_type,
                    compile=parameter.value_type,
                )
            )
            if updated_value_type:
                if isinstance(parameter.value_type, DefaultObjectValueType):
                    # we are going to update
                    parameter = attr.evolve(
                        parameter,
                        value_type=updated_value_type,
                        load_on_build=updated_value_type.load_on_build,
                    )
                    message = "%s: updating parameter with the runtime info" % (message)
                # warn anyway
                warnings.append(message)

    except Exception as ex:
        # we don't want to fail user code on failed value discovery
        # we only print message from "friendly exception" and show real stack
        logger.exception("Failed to discover runtime for %s", parameter)

    try:
        p_val = parameter.calc_init_value(value)
    except Exception as ex:
        raise parameter.parameter_exception(
            "calculate value from '%s'" % safe_string(value, 100), ex=ex
        )

    # we need to break strong reference between tasks
    # otherwise we will have pointer from task to another task
    # if p_val is task, that's ok, but let minimize the risk by patching cf_value
    if isinstance(value, _TaskParamContainer):
        cf_value.value = str(cf_value)

    try:
        if p_val is not None and not isinstance(p_val, Target):
            parameter.validate(p_val)
    except Exception as ex:
        raise parameter.parameter_exception(
            "validate value='%s'" % safe_string(p_val), ex=ex
        )

    p_value = ParameterValue(
        parameter=parameter,
        source=cf_value.source,
        source_value=cf_value.value,
        value=p_val,
        parsed=cf_value.require_parse,
        warnings=warnings + cf_value.warnings,
    )

    return p_value


def infer_parameter_value_type(parameter, value):
    warnings = []
    runtime_value_type = get_types_registry().get_value_type_of_obj(value)
    if not runtime_value_type:
        runtime_value_type = DefaultObjectValueType()
    if runtime_value_type != parameter.value_type:
        if not isinstance(parameter.value_type, DefaultObjectValueType):
            message = (
                "{parameter}: type of the value at runtime '{runtime}'"
                " doesn't match user defined type '{compile}'".format(
                    parameter=parameter,
                    runtime=runtime_value_type,
                    compile=parameter.value_type,
                )
            )
            warnings.append(message)

        parameter = attr.evolve(
            parameter,
            value_type=runtime_value_type,
            value_type_defined=parameter.value_type,
            load_on_build=runtime_value_type.load_on_build,
        )
    return parameter, warnings


def _add_description(base, extra):
    base = base or ""

    if base:
        base += " "
    base += extra



import logging

from dbnd._core.current import get_databand_context
from dbnd.api.jobs import is_job_exists
from dbnd.api.query_params import build_query_api_params, create_filters_builder
from dbnd.api.scheduler import is_scheduled_job_exists
from dbnd.api.shared_schemas.alerts_def_schema import AlertDefsSchema


logger = logging.getLogger(__name__)


def not_none_dict(d):
    return {key: value for key, value in d.items() if value is not None}


def _build_alert(
    job_name,
    job_id,
    task_name,
    uid,
    alert_class,
    severity,
    operator,
    value,
    user_metric,
):
    """build alert for job"""

    return not_none_dict(
        {
            "type": alert_class,
            "job_name": job_name,
            "job_id": job_id,
            "task_name": task_name,
            "uid": uid,
            "operator": operator,
            "value": str(value),
            "severity": severity,
            "user_metric": user_metric,
        }
    )


def run_if_job_exists(func):
    def inner(job_name, *args, **kwargs):
        if is_job_exists(job_name) or is_scheduled_job_exists(job_name):
            return func(job_name, *args, **kwargs)
        else:
            raise LookupError("Job named '%s' is not found" % job_name)

    return inner


def _post_alert(client, alert):
    alert_uid = client.api_request(endpoint="alert_defs", data=alert, method="POST")
    return alert_uid


def create_alert(
    job_name,
    job_id,
    task_name,
    uid,
    alert_class,
    severity,
    operator,
    value,
    user_metric,
):
    """add alert for existing job or scheduled job"""
    alert = _build_alert(
        job_name,
        job_id,
        task_name,
        uid,
        alert_class,
        severity,
        operator,
        value,
        user_metric,
    )
    alert_def_uid = _post_alert(get_databand_context().databand_api_client, alert)
    return alert_def_uid


@run_if_job_exists
def list_job_alerts(job_name):
    return get_alerts_filtered(job_name=job_name)


def _get_alerts(client, query_api_args):
    endpoint = "?".join(["alert_defs", query_api_args])
    schema = AlertDefsSchema(strict=False)
    response = client.api_request(endpoint=endpoint, data="", method="GET")
    return schema.load(data=response["data"], many=True).data


build_alerts_filter = create_filters_builder(
    job_name=("job_name", "eq"),
    severity=("severity", "eq"),
    alert_type=("type", "eq"),
    task_name=("task_name", "eq"),
    scheduled_job_uid=("scheduled_job_uid", "eq"),
    alert_uid=("uid", "eq"),
    custom_name=("custom_name", "eq"),
)


def get_alerts_filtered(**kwargs):
    """get alerts by filters"""
    query_params = build_query_api_params(filters=build_alerts_filter(**kwargs))
    return _get_alerts(get_databand_context().databand_api_client, query_params)


def delete_alerts(uids):
    """delete alerts by uids"""
    get_databand_context().databand_api_client.api_request(
        endpoint="alert_defs/delete", data=uids, method="POST"
    )



import datetime
import uuid

import pytz
import six

from dbnd._core.utils.basics.memoized import cached
from dbnd._vendor import pendulum


def get_uuid():
    # TODO: obfuscate getnode() - mac address part
    return uuid.uuid1()


NAMESPACE_DBND = uuid.uuid5(uuid.NAMESPACE_DNS, "databand.ai")
NAMESPACE_DBND_JOB = uuid.uuid5(NAMESPACE_DBND, "job")
NAMESPACE_DBND_RUN = uuid.uuid5(NAMESPACE_DBND, "run")
NAMESPACE_DBND_TASK_DEF = uuid.uuid5(NAMESPACE_DBND, "task_definition")


def get_task_def_uid(dag_id, task_id, code_hash):
    return uuid.uuid5(
        NAMESPACE_DBND_TASK_DEF, "{}.{}.{}".format(dag_id, task_id, code_hash)
    )


def get_task_run_uid(run_uid, dag_id, task_id):
    return uuid.uuid5(run_uid, "{}.{}".format(dag_id, task_id))


def get_task_run_attempt_uid(run_uid, dag_id, task_id, try_number):
    return uuid.uuid5(run_uid, "{}.{}:{}".format(dag_id, task_id, try_number))


def get_job_run_uid(airflow_instance_uid, dag_id, execution_date):
    if isinstance(execution_date, six.string_types):
        execution_date = pendulum.parse(execution_date)
    if isinstance(execution_date, datetime.datetime):
        execution_date = (
            execution_date.replace(microsecond=0).astimezone(pytz.utc).isoformat()
        )
    if airflow_instance_uid is None:
        return uuid.uuid5(NAMESPACE_DBND_RUN, "{}:{}".format(dag_id, execution_date))
    else:
        return uuid.uuid5(
            NAMESPACE_DBND_RUN,
            "{}:{}:{}".format(airflow_instance_uid, dag_id, execution_date),
        )


def get_job_uid(airflow_server_info_uid, dag_id):
    if airflow_server_info_uid:
        return uuid.uuid5(
            NAMESPACE_DBND_JOB, "{}:{}".format(airflow_server_info_uid, dag_id)
        )
    else:
        return uuid.uuid5(NAMESPACE_DBND_JOB, dag_id)


@cached()
def get_airflow_instance_uid():
    """ used to distinguish between jobs of different airflow instances """
    import airflow

    db_url = airflow.settings.Session.bind.engine.url
    db_str = "{}:{}/{}".format(db_url.host, db_url.port, db_url.database)
    airflow_instance_uid = uuid.uuid5(uuid.NAMESPACE_URL, db_str)
    return str(airflow_instance_uid)



import sys

import six


if six.PY3:
    import contextlib
    from collections.abc import Callable
else:
    import contextlib2 as contextlib
    from collections import Callable


def qualname_func(func):
    if six.PY3:
        return func.__qualname__
    return "%s.%s" % (func.__module__, func.__name__)


try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

try:
    import_errors = (ImportError, ModuleNotFoundError)
except Exception:
    # we are python2
    import_errors = (ImportError,)

try:
    from thread import get_ident
except ImportError:
    try:
        from _thread import get_ident
    except ImportError:
        from _dummy_thread import get_ident


def fix_sys_path_str():
    if six.PY2:
        # fix path from "non" str values, otherwise we fail on py2
        sys.path = [str(p) if type(p) != str else p for p in sys.path]


__all__ = [
    "contextlib",
    "qualname_func",
    "import_errors",
    "Callable",
    "get_ident",
    "fix_sys_path_str",
]



import io
import os

import six

from dbnd._core.settings import LoggingConfig
from dbnd._core.utils.string_utils import merge_dbnd_and_spark_logs
from dbnd._vendor.termcolor import colored


MERGE_MSG_COLOR = "yellow"

if six.PY2:
    from itertools import izip_longest as zip_longest
else:
    from itertools import zip_longest


LINE_BRAKE = "\r\n"
MERGE_MSG = "\r\n...\r\n\r\nThe log body is truncated by databand, fetched {head_size} bytes for the `head` and {tail_size} bytes for the `tail` from the whole {file_size} bytes of the file.\r\nFor full log access the log path {log_path} .\r\nControl the log preview length with LoggingConfig.preview_head_bytes and LoggingConfig.preview_tail_bytes \r\n\r\n...\r\n"
EMPTY_LOG_MSG = "Log truncated to empty"


def read_dbnd_log_preview(log_path, spark_log_path=None):
    """
    Reads the log_path using the current logging config
    """
    logger_config = LoggingConfig.current()
    max_head_size = logger_config.preview_head_bytes
    max_tail_size = logger_config.preview_tail_bytes

    # there is not need to run any of the logic if we don't need to read anything
    if max_head_size == 0 and max_tail_size == 0:
        return EMPTY_LOG_MSG

    if spark_log_path:
        # we need to read and than merge the content of dbnd log and spark long
        # using half of the max_bytes for each of the logs to make sure we are not exceeding the max after merging
        parts = merge_read_log_files(
            log_path, spark_log_path, max_head_size // 2, max_tail_size // 2,
        )

    else:
        # safely reading a single log file
        parts = read_head_and_tail(log_path, max_head_size, max_tail_size)

    # check if all the parts contains information
    if not any(parts):
        # all the parts are truncated
        return EMPTY_LOG_MSG

    # join each part with new lines
    joined_parts = [LINE_BRAKE.join(part) for part in parts]

    # building the merge message between head and tail
    merge_msg = colored(
        MERGE_MSG.format(
            head_size=max_head_size,
            tail_size=max_tail_size,
            file_size=os.path.getsize(log_path),
            log_path=log_path,
        ),
        MERGE_MSG_COLOR,
    )

    return merge_msg.join(joined_parts)


def read_head_and_tail(path, head_size, tail_size):
    """
    Safe reading of the file:
    * Reads the whole file if its size is less than head_size + tail_size
    * Otherwise: Reads the head and the tail of the file separately.

    Parameters:
        path (str): the location of the file to read.
        head_size (int): max length of the the head part in bytes.
        tail_size (int): max length of the the tail part in bytes.

    Returns:
        A tuple with one (whole file) or two (head and tail) list of lines.
    """
    if not (_is_non_negative_int(head_size) and _is_non_negative_int(tail_size)):
        raise ValueError("tail and head max size has to be non negative integers")

    file_size = os.path.getsize(path)
    with open(path, "rb") as f:
        if head_size + tail_size < file_size:
            # readlines(0) returns all the lines but in our case we need no lines
            head_content = f.readlines(head_size) if head_size else []

            if f.seekable():
                f.seek(-tail_size, io.SEEK_END)
            # removing the first line cause it may not be a complete line
            tail_content = f.readlines(tail_size)[1:]

            return _decode_lines(head_content), _decode_lines(tail_content)

        else:
            return (_decode_lines(f.readlines()),)


def merge_read_log_files(dbnd_log_path, spark_log_path, head_size, tail_size):
    """
    Handle the safe reading and merging of two log files.
    """
    # we create tuples of parts (heads together and tails together).
    # if there is no tail for only one of them will have : [(dbnd_head, spark_head), (dbnd_head, [])]
    parts = zip_longest(
        read_head_and_tail(dbnd_log_path, head_size, tail_size),
        read_head_and_tail(spark_log_path, head_size, tail_size),
        fillvalue=[],
    )

    # merging each part (heads, tails) separately
    merged_parts = []
    for part in parts:
        merged = merge_dbnd_and_spark_logs(*part)
        merged_parts.append(merged)

    return merged_parts


# some helper function to make the code cleaner
def _decode(s, encoding="utf-8", errors="strict"):
    """
    helper to transfer decode method to a function
    """
    return s.decode(encoding, errors)


def _decode_lines(lines):
    """
    helper to decodes multiple lines of strings
    """
    return [line.rstrip(LINE_BRAKE) for line in map(_decode, lines)]


def _is_non_negative_int(value):
    return isinstance(value, int) and value >= 0



import datetime
import logging
import typing

from dbnd._core.constants import DbndTargetOperationStatus, DbndTargetOperationType
from dbnd._core.tracking.backends.abstract_tracking_store import TrackingStore
from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.tracking.tracking_info_convertor import TrackingInfoBuilder
from dbnd._core.utils import json_utils
from dbnd._core.utils.timezone import utcnow
from dbnd.api.tracking_api import LogTargetArgs, TaskRunAttemptUpdateArgs
from targets.value_meta import ValueMeta


if typing.TYPE_CHECKING:
    from typing import List, Optional, Iterable
    from uuid import UUID

    from dbnd._core.constants import TaskRunState
    from dbnd._core.task_run.task_run import TaskRun
    from dbnd._core.task_run.task_run_error import TaskRunError

logger = logging.getLogger(__name__)


class TrackingStoreThroughChannel(TrackingStore):
    """Track data to Tracking API"""

    def __init__(self, channel):
        # type: (TrackingChannel) -> None
        self.channel = channel

    def init_scheduled_job(self, scheduled_job, update_existing):
        return self._m(
            self.channel.init_scheduled_job,
            scheduled_job_args=scheduled_job,
            update_existing=update_existing,
        )

    def init_run(self, run):
        init_args = TrackingInfoBuilder(run).build_init_args()
        return self.init_run_from_args(init_args=init_args)

    def init_run_from_args(self, init_args):
        return self._m(self.channel.init_run, init_args=init_args)

    def add_task_runs(self, run, task_runs):
        task_runs_info = TrackingInfoBuilder(run).build_task_runs_info(
            task_runs=task_runs, dynamic_task_run_update=True
        )
        return self._m(
            self.channel.add_task_runs,
            task_runs_info=task_runs_info,
            source=run.source,
        )

    def set_run_state(self, run, state, error=None, timestamp=None):
        return self._m(
            self.channel.set_run_state,
            run_uid=run.run_uid,
            state=state,
            timestamp=timestamp,
        )

    def set_task_reused(self, task_run):
        return self._m(
            self.channel.set_task_reused,
            task_run_uid=task_run.task_run_uid,
            task_outputs_signature=task_run.task.task_meta.task_outputs_signature,
        )

    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        # type: (TaskRun, TaskRunState, TaskRunError, datetime.datetime) -> None
        return self._m(
            self.channel.update_task_run_attempts,
            task_run_attempt_updates=[
                TaskRunAttemptUpdateArgs(
                    task_run_uid=task_run.task_run_uid,
                    task_run_attempt_uid=task_run.task_run_attempt_uid,
                    state=state,
                    error=error.as_error_info() if error else None,
                    timestamp=timestamp or utcnow(),
                )
            ],
        )

    def set_task_run_states(self, task_runs):
        return self._m(
            self.channel.update_task_run_attempts,
            task_run_attempt_updates=[
                TaskRunAttemptUpdateArgs(
                    task_run_uid=task_run.task_run_uid,
                    task_run_attempt_uid=task_run.task_run_attempt_uid,
                    state=task_run.task_run_state,
                    timestamp=utcnow(),
                )
                for task_run in task_runs
            ],
        )

    def set_unfinished_tasks_state(self, run_uid, state):
        return self._m(
            self.channel.set_unfinished_tasks_state,
            run_uid=run_uid,
            state=state,
            timestamp=utcnow(),
        )

    def update_task_run_attempts(self, task_run_attempt_updates):
        return self._m(
            self.channel.update_task_run_attempts,
            task_run_attempt_updates=task_run_attempt_updates,
        )

    def save_task_run_log(self, task_run, log_body, local_log_path=None):
        return self._m(
            self.channel.save_task_run_log,
            task_run_attempt_uid=task_run.task_run_attempt_uid,
            log_body=log_body,
            local_log_path=local_log_path,
        )

    def save_external_links(self, task_run, external_links_dict):
        return self._m(
            self.channel.save_external_links,
            task_run_attempt_uid=task_run.task_run_attempt_uid,
            external_links_dict=external_links_dict,
        )

    def log_target(
        self,
        task_run,
        target,
        target_meta,  # type: ValueMeta
        operation_type,  # type: DbndTargetOperationType
        operation_status,  # type: DbndTargetOperationStatus
        param_name=None,  # type: Optional[str]
        task_def_uid=None,  # type: Optional[UUID]
    ):
        data_schema = (
            json_utils.dumps(target_meta.data_schema)
            if target_meta.data_schema is not None
            else None
        )
        target_info = LogTargetArgs(
            run_uid=task_run.run.run_uid,
            task_run_uid=task_run.task_run_uid,
            task_run_name=task_run.job_name,
            task_run_attempt_uid=task_run.task_run_attempt_uid,
            task_def_uid=task_def_uid,
            param_name=param_name,
            target_path=str(target),
            operation_type=operation_type,
            operation_status=operation_status,
            value_preview=target_meta.value_preview,
            data_dimensions=target_meta.data_dimensions,
            data_schema=data_schema,
            data_hash=target_meta.data_hash,
        )
        res = self.log_targets(targets_info=[target_info])
        if target_meta.histograms:
            self.log_histograms(task_run, param_name, target_meta, utcnow())
        return res

    def log_targets(self, targets_info):  # type: (List[LogTargetArgs]) -> None
        return self._m(self.channel.log_targets, targets_info=targets_info)

    def log_histograms(self, task_run, key, value_meta, timestamp):
        value_meta_metrics = value_meta.build_metrics_for_key(key)
        if value_meta_metrics["histograms"]:
            self.log_metrics(
                task_run=task_run, metrics=value_meta_metrics["histograms"]
            )

    def log_metrics(self, task_run, metrics):
        # type: (TaskRun, Iterable[Metric]) -> None
        metrics_info = [
            {
                "task_run_attempt_uid": task_run.task_run_attempt_uid,
                "metric": metric,
                "source": metric.source,
            }
            for metric in metrics
        ]
        return self._m(self.channel.log_metrics, metrics_info=metrics_info,)

    def log_artifact(self, task_run, name, artifact, artifact_target):
        return self._m(
            self.channel.log_artifact,
            task_run_attempt_uid=task_run.task_run_attempt_uid,
            name=name,
            path=artifact_target.path,
        )

    def heartbeat(self, run_uid):
        return self._m(self.channel.heartbeat, run_uid=run_uid)

    def save_airflow_task_infos(self, airflow_task_infos, source, base_url):
        return self._m(
            self.channel.save_airflow_task_infos,
            airflow_task_infos=airflow_task_infos,
            source=source,
            base_url=base_url,
        )

    def save_airflow_monitor_data(
        self, airflow_monitor_data, airflow_base_url, last_sync_time
    ):
        return self._m(
            self.channel.save_airflow_monitor_data,
            airflow_export_data=airflow_monitor_data,
            airflow_base_url=airflow_base_url,
            last_sync_time=last_sync_time,
        )

    def _m(self, channel_call, **req_kwargs):
        """
        Marshall and call channel function
        :return:
        """
        req_schema = self.channel.get_schema_by_handler_name(channel_call.__name__)
        marsh = req_schema.dump(req_kwargs)
        resp = channel_call(marsh.data)
        # if resp_schema and resp:
        #     resp = resp_schema.load(resp)
        return resp

    def is_ready(self):
        return self.channel.is_ready()

    def __str__(self):
        return "TrackingStoreThroughChannel with channel=%s" % (str(self.channel),)

    @staticmethod
    def build_with_disabled_channel():
        from dbnd._core.tracking.backends.channels.tracking_disabled_channel import (
            DisabledTrackingChannel,
        )

        return TrackingStoreThroughChannel(channel=DisabledTrackingChannel())

    @staticmethod
    def build_with_console_debug_channel():
        from dbnd._core.tracking.backends.channels.tracking_debug_channel import (
            ConsoleDebugTrackingChannel,
        )

        return TrackingStoreThroughChannel(channel=ConsoleDebugTrackingChannel())

    @staticmethod
    def build_with_web_channel():
        from dbnd._core.tracking.backends.channels.tracking_web_channel import (
            TrackingWebChannel,
        )

        return TrackingStoreThroughChannel(channel=TrackingWebChannel())

    @staticmethod
    def build_with_proto_web_channel():
        from dbnd._core.tracking.backends.channels.tracking_proto_web_channel import (
            TrackingProtoWebChannel,
        )

        return TrackingStoreThroughChannel(channel=TrackingProtoWebChannel())



from __future__ import absolute_import

import base64
import logging

from io import BytesIO

from matplotlib import figure

from dbnd._vendor import fast_hasher
from targets.values.builtins_values import DataValueType


logger = logging.getLogger(__name__)


class MatplotlibFigureValueType(DataValueType):
    type = figure.Figure
    config_name = "matplotlib_figure"

    def to_signature(self, x):
        return fast_hasher.hash(x)

    def to_preview(self, value, preview_size):
        try:
            figure_bytes = BytesIO()
            value.savefig(figure_bytes, format="png")
            figure_bytes.seek(0)
            preview = base64.b64encode(figure_bytes.read())
            if len(preview) > preview_size:
                return (
                    ".png size %s is bigger than %s bytes, change '[feature]/log_value_preview_max_len'"
                    % (len(preview), preview_size)
                )
            return preview
        except Exception as ex:
            logger.warning(
                "Failed to calculate preview for matplotlib figure: %s", str(ex)
            )
            return None



import os
import typing

from typing import Optional

from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.current import try_get_current_task_run
from dbnd._vendor.colorlog import logging


if typing.TYPE_CHECKING:
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


def try_get_or_create_task_run():
    # type: ()-> Optional[TaskRun]
    task_run = try_get_current_task_run()
    if task_run:
        return task_run

    from dbnd._core.configuration.environ_config import DBND_TASK_RUN_ATTEMPT_UID

    tra_uid = os.environ.get(DBND_TASK_RUN_ATTEMPT_UID)
    if tra_uid:
        return _get_task_run_mock(tra_uid)

    from dbnd._core.inplace_run.inplace_run_manager import try_get_inplace_task_run

    return try_get_inplace_task_run()


def _get_task_run_mock(tra_uid):
    """
    We need better implementation for this,
    currently in use only for spark
    """
    try:
        from dbnd._core.task_run.task_run_tracker import TaskRunTracker

        task_run = TaskRunMock(tra_uid)
        from dbnd import config
        from dbnd._core.settings import CoreConfig

        with config(
            {CoreConfig.tracker_raise_on_error: False}, source="on_demand_tracking"
        ):
            with new_dbnd_context(
                name="fast_dbnd_context", autoload_modules=False
            ) as fast_dbnd_ctx:
                trt = TaskRunTracker(task_run, fast_dbnd_ctx.tracking_store)
                task_run.tracker = trt
                return task_run
    except Exception:
        logger.info("Failed during dbnd inplace tracking init.", exc_info=True)
        return None


class TaskRunMock(object):
    def __init__(self, task_run_attempt_uid):
        self.task_run_attempt_uid = task_run_attempt_uid

    def __getattr__(self, item):
        return self.__dict__.get(item, self)



from dbnd._core.tracking.backends.abstract_tracking_store import TrackingStore
from dbnd._core.tracking.backends.tracking_store_channels import (
    TrackingStoreThroughChannel,
)
from dbnd._core.tracking.backends.tracking_store_composite import CompositeTrackingStore
from dbnd._core.tracking.backends.tracking_store_console import ConsoleStore
from dbnd._core.tracking.backends.tracking_store_file import FileTrackingStore
from dbnd._core.tracking.backends.tracking_store_tf_summary import TbSummaryFileStore


__all__ = [
    TrackingStore,
    TrackingStoreThroughChannel,
    CompositeTrackingStore,
    ConsoleStore,
    FileTrackingStore,
    TbSummaryFileStore,
]



from dbnd._core.task_build.task_const import _SAME_AS_PYTHON_MODULE
from dbnd._core.task_build.task_registry import get_task_registry


def namespace(namespace=None, scope=""):
    """
    Call to set namespace of tasks declared after the call.

    It is often desired to call this function with the keyword argument
    ``scope=__name__``.

    The ``scope`` keyword makes it so that this call is only effective for task
    classes with a matching [*]_ ``__module__``. The default value for
    ``scope`` is the empty string, which means all classes. Multiple calls with
    the same scope simply replace each other.

    The namespace of a :py:class:`Task` can also be changed by specifying the property
    ``task_namespace``.

    .. code-block:: python

        class Task2(dbnd.Task):
            task_namespace = 'namespace2'

    This explicit setting takes priority over whatever is set in the
    ``namespace()`` method, and it's also inherited through normal python
    inheritence.

    There's no equivalent way to set the ``task_family``.

    .. [*] When there are multiple levels of matching module scopes like
           ``a.b`` vs ``a.b.c``, the more specific one (``a.b.c``) wins.
    .. seealso:: The new and better scaling :py:func:`auto_namespace`
    """
    get_task_registry().register_namespace(scope=scope, namespace=namespace or "")


def auto_namespace(scope=""):
    """
    Same as :py:func:`namespace`, but instead of a constant namespace, it will
    be set to the ``__module__`` of the task class. This is desirable for these
    reasons:

     * Two tasks with the same name will not have conflicting task families
     * It's more pythonic, as modules are Python's recommended way to
       do namespacing.
     * It's traceable. When you see the full name of a task, you can immediately
       identify where it is defined.

    We recommend calling this function from your package's outermost
    ``__init__.py`` file. The file contents could look like this:

    .. code-block:: python

        import dbnd

        databand.auto_namespace(scope=__name__)

    To reset an ``auto_namespace()`` call, you can use
    ``namespace(scope='my_scope'``).  But this will not be
    needed (and is also discouraged) if you use the ``scope`` kwarg.
    """
    namespace(namespace=_SAME_AS_PYTHON_MODULE, scope=scope)



import attr

from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.utils.task_utils import to_targets


@attr.s(hash=False, repr=False, str=False)
class _TargetParameter(ParameterDefinition):  # extracted for generics support
    """
      Parameter whose value is a ```target```.

      In the task definition, use

      .. code-block:: python

          class MyTask(dbnd.Task):
              some_input = data

              def run(self):
                    self.some_input.open('r')

      """

    def normalize(self, x):
        # can not move to value_type, we need target_config
        return to_targets(x, from_string_kwargs=dict(config=self.target_config))



from __future__ import absolute_import

import numpy

from dbnd._vendor import fast_hasher
from targets.values.builtins_values import DataValueType


class NumpyArrayValueType(DataValueType):
    type = numpy.ndarray
    config_name = "numpy_ndarray"
    support_merge = True

    def to_signature(self, x):
        shape = "[%s]" % (",".join(map(str, x.shape)))
        return "%s:%s" % (shape, fast_hasher.hash(x))

    def merge_values(self, *values, **kwargs):
        return numpy.concatenate(values)



import datetime
import functools
import json
import operator

from collections import OrderedDict
from typing import Mapping
from uuid import UUID

import dbnd._vendor.hjson as hjson

from dbnd._core.utils.platform import windows_compatible_mode
from dbnd._core.utils.traversing import DatabandDict
from dbnd._core.utils.type_check_utils import is_instance_by_class_name


class _FrozenOrderedDict(Mapping, DatabandDict):
    """
    .. deprecated:: 0.7.3

    It is an immutable wrapper around ordered dictionaries that implements the complete :py:class:`collections.Mapping`
    interface. It can be used as a drop-in replacement for dictionaries where immutability and ordering are desired.
    """

    def __init__(self, *args, **kwargs):
        self.__dict = OrderedDict(*args, **kwargs)
        self.__hash = None

    def __getitem__(self, key):
        return self.__dict[key]

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return repr(self.__dict)

    def __hash__(self):
        if self.__hash is None:
            hashes = map(hash, self.items())
            self.__hash = functools.reduce(operator.xor, hashes, 0)

        return self.__hash

    def get_wrapped(self):
        return self.__dict

    def copy(self):
        return _FrozenOrderedDict(self.__dict)


def json_default(obj, safe=False):
    from dbnd._core.parameter.parameter_definition import ParameterDefinition

    if isinstance(obj, ParameterDefinition):
        return str(obj)

    if isinstance(obj, _FrozenOrderedDict):
        return obj.get_wrapped()

    from targets import Target

    if isinstance(obj, Target):
        return str(obj)

    if isinstance(obj, datetime.datetime):
        return obj.strftime("%Y-%m-%dT%H:%M:%SZ")
    elif isinstance(obj, datetime.date):
        return obj.strftime("%Y-%m-%d")

    if is_instance_by_class_name(obj, "int32") or is_instance_by_class_name(
        obj, "int64"
    ):
        return str(obj)

    if isinstance(obj, UUID):
        return str(obj)
    if safe:
        return str(obj)

    raise TypeError(repr(obj) + " is not JSON serializable")


def dumps(obj, default=json_default, sort_keys=True, **kwargs):
    return json.dumps(obj, default=default, sort_keys=sort_keys, **kwargs)


def loads(s, **kwargs):
    if windows_compatible_mode:
        if s and s[0] in ["{", "["]:
            try:
                return loads_canonical(s, **kwargs)
            except Exception:
                return hjson.loads(s, **kwargs)
        return s
    return hjson.loads(s, **kwargs)


def loads_canonical(s, **kwargs):
    return json.loads(s, **kwargs)


def dumps_canonical(obj, separators=(",", ":"), sort_keys=True, default=json_default):
    return json.dumps(obj, separators=separators, sort_keys=sort_keys, default=default)


def dumps_safe(
    obj,
    separators=(",", ":"),
    sort_keys=True,
    default=functools.partial(json_default, safe=True),
):
    return dumps_canonical(
        obj=obj, separators=separators, sort_keys=sort_keys, default=default
    )


def convert_to_safe_types(obj):
    return loads(dumps_safe(obj))



import errno
import logging
import os


logger = logging.getLogger(__name__)

try:
    from urlparse import urlsplit
except ImportError:
    from urllib.parse import urlsplit


def trailing_slash(path):
    # I suppose one day schema-like paths, like
    # file:///path/blah.txt?params=etc can be parsed too
    return path[-1] if path[-1] in r"\/" else ""


def no_trailing_slash(path):
    if not path:
        return path
    return path[:-1] if path[-1] in r"\/" else path


def safe_mkdirs(path, mode):
    """
    safe version of make dir
    """

    o_umask = os.umask(0)
    try:
        os.makedirs(path, mode)
    except OSError as e:
        if e.errno != errno.EEXIST:
            logger.error("Could not create dir: %s. %s", path, e)
            raise
    finally:
        os.umask(o_umask)


def path_to_bucket_and_key(path):
    (scheme, netloc, path, _, _) = urlsplit(str(path))
    path_without_initial_slash = path[1:]
    return netloc, path_without_initial_slash
